/*
 * Copyright (c) 2005, The Regents of the University of California, through
 * Lawrence Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy). All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * (2) Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * (3) Neither the name of the University of California, Lawrence Berkeley
 * National Laboratory, U.S. Dept. of Energy nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * You are under no obligation whatsoever to provide any bug fixes, patches, or
 * upgrades to the features, functionality or performance of the source code
 * ("Enhancements") to anyone; however, if you choose to make your Enhancements
 * available either publicly, or directly to Lawrence Berkeley National
 * Laboratory, without imposing a separate written license agreement for such
 * Enhancements, then you hereby grant the following license: a non-exclusive,
 * royalty-free perpetual license to install, use, modify, prepare derivative
 * works, incorporate into other computer software, distribute, and sublicense
 * such enhancements or derivative works thereof, in binary and source code
 * form.
 */
package nux.xom.io;

import java.io.IOException;
import java.io.OutputStream;

import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

import nu.xom.Attribute;
import nu.xom.Comment;
import nu.xom.DocType;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.IllegalAddException;
import nu.xom.Node;
import nu.xom.ProcessingInstruction;
import nu.xom.Text;

/**
 * TODO; EXPERIMENTAL; A {@link StreamingSerializer} implementation that
 * delegates to an underlying STAX {@link javax.xml.stream.XMLStreamWriter}.
 * 
 * @author whoschek.AT.lbl.DOT.gov
 * @author $Author: hoschek3 $
 * @version $Revision: 1.5 $, $Date: 2005/11/30 06:28:16 $
 */
final class StreamingStaxSerializer implements StreamingSerializer {
	
	/** The underlying STAX writer to delegate to. */
	private final XMLStreamWriter writer;
	
	/** STAX does not perform wellformedness checks, so we do it ourselves. */
	private final StreamingVerifier verifier = new StreamingVerifier();
	
	/**
	 * STAX impls tend to do a poor job wrt. eliminating unnecessary namespace
	 * declarations, so we do it ourselves.
	 */
	private final NamespacesInScope namespaces = new NamespacesInScope();
	
	private static final boolean DEBUG = false;
		
	
	/**
	 * Constructs a new object that writes to the given underlying STAX
	 * XMLStreamWriter.
	 * 
	 * @param writer
	 *            the underlying XMLStreamWriter to write to
	 */
	StreamingStaxSerializer(XMLStreamWriter writer) {
		if (writer == null) 
			throw new IllegalArgumentException("XMLStreamWriter must not be null");

		this.writer = writer;
	}
	
	StreamingStaxSerializer(OutputStream out, String encoding) throws XMLStreamException {
		this(XMLOutputFactory.newInstance().createXMLStreamWriter(out, encoding));
	}
	
	private void reset() {
		verifier.reset();
		namespaces.reset();
	}

	/** {@inheritDoc} */
	public void flush() throws IOException {
		try {
			writer.flush();
		} catch (XMLStreamException e) {
			wrapException(e);
		}
	}
	
	/** {@inheritDoc} */
	public void writeStartTag(Element elem) throws IOException {
		verifier.writeStartTag(elem);
		namespaces.push();
		try {
			writer.writeStartElement(
				elem.getNamespacePrefix(), 
				elem.getLocalName(), 
				elem.getNamespaceURI());
		} catch (XMLStreamException e) {
			wrapException(e);
		}
		
		writeAttributes(elem);		
		writeNamespaceDeclarations(elem);
	}
	
	private void writeNamespaceDeclarations(Element elem) throws IOException {
		// TODO: could be implemented more efficiently via 
		// elem.getAdditionalNamespaceDeclarations + attributes iter + elem.getNamespaceURI
		int count = elem.getNamespaceDeclarationCount();
		for (int i=0; i < count; i++) {
			String prefix = elem.getNamespacePrefix(i);
			String uri = elem.getNamespaceURI(prefix);
			if (namespaces.addIfAbsent(prefix, uri)) {
				try {
					writer.writeNamespace(prefix, uri);
				} catch (XMLStreamException e) {
					wrapException(e);
				}
			}
		}
	}
	
	private void writeAttributes(Element elem) throws IOException {
		int count = elem.getAttributeCount();
		for (int i=0; i < count; i++) {
			Attribute attr = elem.getAttribute(i);
			try {
				writer.writeAttribute(
					attr.getNamespacePrefix(), 
					attr.getNamespaceURI(), 
					attr.getLocalName(), 
					attr.getValue());
			} catch (XMLStreamException e) {
				wrapException(e);
			}
		}
	}
	
	/** {@inheritDoc} */
	public void writeEndTag() throws IOException {
		verifier.writeEndTag();
		try {
			writer.writeEndElement();
		} catch (XMLStreamException e) {
			wrapException(e);
		}
		namespaces.pop();
	}
	
	/** {@inheritDoc} */
	public void write(Document doc) throws IOException {
		writeXMLDeclaration();
		for (int i = 0; i < doc.getChildCount(); i++) {
			writeChild(doc.getChild(i));
		}
		writeEndDocument();
	}
	
	/** {@inheritDoc} */
	public void write(Element element) throws IOException {
		if (DEBUG) System.err.println("writing:" + element.getQualifiedName());
		writeStartTag(element);
		for (int i=0; i < element.getChildCount(); i++) {
			writeChild(element.getChild(i));
		}
		writeEndTag();
	}
	
	/** {@inheritDoc} */
	public void write(Text text) throws IOException {
		verifier.write(text);
		String value = text.getValue();
		if (value.length() > 0) {
			try {
				writer.writeCharacters(value);
			} catch (XMLStreamException e) {
				wrapException(e);
			}
		}
	}
	
	/** {@inheritDoc} */
	public void write(Comment comment) throws IOException {
		verifier.write(comment);
		try {
			writer.writeComment(comment.getValue());
		} catch (XMLStreamException e) {
			wrapException(e);
		}
	}
	
	/** {@inheritDoc} */
	public void write(ProcessingInstruction instruction) throws IOException {
		verifier.write(instruction);
		try {
			writer.writeProcessingInstruction(
					instruction.getTarget(), 
					instruction.getValue());
		} catch (XMLStreamException e) {
			wrapException(e);
		}
	}
	
	/** {@inheritDoc} */
	public void write(DocType docType) throws IOException {
		verifier.write(docType);
		try {
			writer.writeDTD(docType.toXML()); // FIXME ???
		} catch (XMLStreamException e) {
			wrapException(e);
		}
	}
	
	/** {@inheritDoc} */
	public void writeEndDocument() throws IOException {
		for (int i=verifier.depth(); --i >= 0; ) { // close all remaining open tags 
			writeEndTag();
		}
		verifier.writeEndDocument();
		try {
			writer.writeEndDocument();
		} catch (XMLStreamException e) {
			wrapException(e);
		}
		flush();
		reset();
	}

	/** {@inheritDoc} */
	public void writeXMLDeclaration() throws IOException {
		verifier.writeXMLDeclaration(); // do sanity check
		reset();
		verifier.writeXMLDeclaration(); // sets hasXMLDeclaration = true
		try {
			writer.writeStartDocument();
		} catch (XMLStreamException e) {
			wrapException(e);
		}
	}
	
	private void writeChild(Node node) throws IOException {
		if (node instanceof Element) {
			write (((Element) node));
		} else if (node instanceof Text) {
			write(((Text) node));
		} else if (node instanceof Comment) {
			write(((Comment) node));
		} else if (node instanceof ProcessingInstruction) {
			write(((ProcessingInstruction) node));
		} else if (node instanceof DocType) {
			write(((DocType) node));
		} else {
			throw new IllegalAddException("Cannot write node type: " + node);
		}
	}

	private void wrapException(Throwable e) throws IOException {
		throw new XMLStreamIOException(e);
	}
	
	// TODO: make this public?
	private static final class XMLStreamIOException extends IOException {
		private XMLStreamIOException(Throwable cause) {
			initCause(cause);
		}
	}
			
}