/*
 * Copyright (c) 2005, The Regents of the University of California, through
 * Lawrence Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy). All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * (2) Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * (3) Neither the name of the University of California, Lawrence Berkeley
 * National Laboratory, U.S. Dept. of Energy nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * You are under no obligation whatsoever to provide any bug fixes, patches, or
 * upgrades to the features, functionality or performance of the source code
 * ("Enhancements") to anyone; however, if you choose to make your Enhancements
 * available either publicly, or directly to Lawrence Berkeley National
 * Laboratory, without imposing a separate written license agreement for such
 * Enhancements, then you hereby grant the following license: a non-exclusive,
 * royalty-free perpetual license to install, use, modify, prepare derivative
 * works, incorporate into other computer software, distribute, and sublicense
 * such enhancements or derivative works thereof, in binary and source code
 * form.
 */
package nux.xom.sandbox;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.Arrays;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import net.sf.saxon.Configuration;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.query.StaticQueryContext;
import nu.xom.Builder;
import nu.xom.DocType;
import nu.xom.Document;
import nu.xom.NodeFactory;
import nu.xom.Serializer;
import nu.xom.converters.SAXConverter;
import nux.xom.binary.BinaryXMLCodec;
import nux.xom.io.StreamingSerializer;
import nux.xom.io.StreamingSerializerFactory;
import nux.xom.pool.FileUtil;
import nux.xom.pool.XOMUtil;

import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

/**
 * Tests and benchmarks bnux format against a set of standard textual XML files.
 * Command line options allow to round-trip and check canonical XML results.
 * They also allow to benchmark serialization, deserialization or both, 
 * either via bnux or xom.
 * <p>
 * Example usage: 
 * <pre>
 * java -server nux.xom.sandbox.BinaryXMLTest test bnux 0 1 1 samples/data/*.xml
 * java -server nux.xom.sandbox.BinaryXMLTest ser bnux 0 10000 5 samples/data/periodic.xml
 * java -server nux.xom.sandbox.BinaryXMLTest deser bnux 0 10000 5 samples/data/periodic.xml
 * java -server nux.xom.sandbox.BinaryXMLTest serdeser bnux 0 5000 5 samples/data/periodic.xml
 * java -server nux.xom.sandbox.BinaryXMLTest serdeser xom 0 5000 5 samples/data/periodic.xml
 * </pre>
 * 
 * @author whoschek.AT.lbl.DOT.gov
 * @author $Author: hoschek3 $
 * @version $Revision: 1.2 $, $Date: 2005/11/30 03:54:38 $
 */
public final class BinaryXMLTest {
	
	private BinaryXMLTest() {}
	
	public static void main(final String args[]) throws Exception {
		
		final String cmd = args[0]; // ser|deser|serdeser|test
		final String mode = args[1]; // bnux|xom|saxon|dom|fi
		int compressionLevel = Integer.parseInt(args[2]); // 0..9
		final int iterations = Integer.parseInt(args[3]); // 1..infinity
		final int runs = Integer.parseInt(args[4]); // 1..infinity
		final boolean enablePerformancePatches = "true".equals(
			System.getProperty("nux.xom.tests.BinaryXMLTest.enablePatches", "true"));
		System.out.println("patchesEnabled=" + enablePerformancePatches);
		final boolean testCompressionLevels = false;

		for (int run=0; run < runs; run++) { // make k large to stress test for a couple of hours 
			long time = 0;
			long doneEncoded = 0;
			long done = 0;
			int checksum = 0;
			
			if (enablePerformancePatches) { // init before BinaryXMLCodec
				// temporary (?) performance hack via patch: disable some expensive sanity checks 
				System.setProperty("nu.xom.Verifier.checkPCDATA", "false");
				System.setProperty("nu.xom.Verifier.checkURI", "false");
			}
			
			if (mode.indexOf("sun") >= 0 && mode.indexOf("stax") >= 0) {
				System.setProperty("javax.xml.stream.XMLInputFactory", "com.sun.xml.stream.ZephyrParserFactory");
				System.setProperty("javax.xml.stream.XMLOutputFactory", "com.sun.xml.stream.ZephyrWriterFactory");
				System.setProperty("javax.xml.stream.XMLEventFactory", "com.sun.xml.stream.events.ZephyrEventFactory");
			}
			
			// bnux and XOM
			BinaryXMLCodec codec = new BinaryXMLCodec();
			NodeFactory bnuxFactory = null;
			if (mode.startsWith("bnux")) {
				if (mode.indexOf("NNF") >= 0) {
					bnuxFactory = XOMUtil.getNullNodeFactory();
				}
			}

			Builder builder = new Builder();
//			Builder builder = BuilderPool.GLOBAL_POOL.getBuilder(false);
			if (mode.equals("xom-V")) {
				builder = new Builder(new NodeFactory() {});
			}
			if (mode.equals("xom-NNF")) {
				builder = new Builder(XOMUtil.getNullNodeFactory());
			}
			
			// saxon
			StaticQueryContext context = null;
			Transformer saxonSerializer = null;
			if (mode.equals("saxon")) {
				context = new StaticQueryContext(new Configuration());
				String clazz = "net.sf.saxon.TransformerFactoryImpl";
				System.setProperty("javax.xml.transform.TransformerFactory", clazz);
				saxonSerializer = TransformerFactory.newInstance().newTransformer();
			}
			
			// DOM
			DocumentBuilder domBuilder = null;
			Transformer domSerializer = null;
			if (mode.equals("dom")) {
				DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
				factory.setNamespaceAware(true);
				factory.setAttribute("http://apache.org/xml/features/dom/defer-node-expansion", Boolean.FALSE);
				domBuilder = factory.newDocumentBuilder();
				String clazz = "com.sun.org.apache.xalan.internal.xsltc.trax.TransformerFactoryImpl";
//				String clazz = "org.apache.xalan.processor.TransformerFactoryImpl";
				System.setProperty("javax.xml.transform.TransformerFactory", clazz);
				domSerializer = TransformerFactory.newInstance().newTransformer();
				System.err.println(domSerializer.getClass().getName());
			}
			
			// FastInfoSet
			ContentHandler fiSerializer = null;
			Builder fiBuilder = null;	
			java.lang.reflect.Method fiMethod = null;
			
			XMLStreamReader fistaxReader = null;
			java.lang.reflect.Method fistaxMethod = null;
			if (mode.startsWith("fi")) {
				NodeFactory factory = null;
				if (mode.indexOf("NNF") >= 0) factory = XOMUtil.getNullNodeFactory();
				XMLReader parser = (XMLReader) Class.forName("com.sun.xml.fastinfoset.sax.SAXDocumentParser").newInstance();
				fiBuilder = new Builder(parser, false, factory);
				fiSerializer = (ContentHandler) Class.forName("com.sun.xml.fastinfoset.sax.SAXDocumentSerializer").newInstance();
				
				if (mode.startsWith("fi1")) { // enable "full indexing"
					java.lang.reflect.Method method;
					method = fiSerializer.getClass().getMethod(
							"setAttributeValueSizeLimit", new Class[] {Integer.TYPE});
					method.invoke(fiSerializer, new Object[] {new Integer(Integer.MAX_VALUE)});
					method = fiSerializer.getClass().getMethod(
							"setCharacterContentChunkSizeLimit", new Class[] {Integer.TYPE});
					method.invoke(fiSerializer, new Object[] {new Integer(Integer.MAX_VALUE)});
				}
				
				fiMethod = fiSerializer.getClass().getMethod(
						"setOutputStream", new Class[] { OutputStream.class});
				
				if (mode.indexOf("stax") >= 0) {
					fistaxReader = (XMLStreamReader) Class.forName("com.sun.xml.fastinfoset.stax.StAXDocumentParser").newInstance();
					fistaxMethod = fistaxReader.getClass().getMethod(
							"setInputStream", new Class[] { InputStream.class});
				}
			}
			
			// STAX
			StaxBuilder staxBuilder = null;
			if (mode.indexOf("stax") >= 0) {
				NodeFactory factory = null;
				if (mode.indexOf("NNF") >= 0) factory = XOMUtil.getNullNodeFactory();
				staxBuilder = new StaxBuilder(factory);
			}
			
			for (int j=5; j < args.length; j++) {
				try {
					if (ignore(args[j])) continue;
					File file = new File(args[j]);
					if (file.isDirectory()) continue;
					System.out.println("now processing " + file);
//					System.out.print(".");
					
					// prepare
					// file:/path/to/file --> file:///path/to/file
					String baseURI = file.toURI().toASCIIString();
					if (baseURI.startsWith("file:/")) {
						baseURI = baseURI.substring("file:/".length());
						if (!baseURI.startsWith("//")) baseURI = "//" + baseURI;
						baseURI = "file:/" + baseURI;
					}
//					System.out.println("baseURI: " + baseURI);
					Document doc = new Builder().build(file);
					byte[] data = codec.serialize(doc, compressionLevel);
					if (! cmd.equals("test")) doc = new BinaryXMLCodec().deserialize(data); // use "interned" strings
					byte[] fileData = FileUtil.toByteArray(new FileInputStream(file));
					long fileLength = file.length();
					int encodedSize = 0;
					
					org.w3c.dom.Document domDoc = null;
					if (mode.equals("dom")) {
						domDoc = domBuilder.parse(file);
					}
					
					NodeInfo saxonDoc = null;
					if (mode.equals("saxon")) {
						saxonDoc = context.buildDocument(new StreamSource(new ByteArrayInputStream(fileData)));
					}
	
					if (mode.startsWith("fi")) {
						data = serializeWithFastInfoset(doc, fiSerializer, fiMethod);
					}
	
					// run the benchmark
					long start = System.currentTimeMillis();
					for (int i = 0; i < iterations; i++) {
						try {
							// serialize
							if (cmd.equals("ser") || cmd.equals("serdeser") || cmd.equals("test")) {
								if (mode.startsWith("bnux")) {
									if (mode.indexOf("stream") < 0) {
										data = codec.serialize(doc, compressionLevel);
									} else {
										data = serializeWithStreamingBnux(doc, compressionLevel);
									}
								} else if (mode.startsWith("xom")) {
									if (mode.indexOf("stream") < 0) {
										data = serializeWithXOM(doc);
									} else {
										data = serializeWithStreamingXOM(doc);
									}
								} else if (mode.equals("saxon")) {
									ByteArrayOutputStream out = new ByteArrayOutputStream();
									saxonSerializer.transform(saxonDoc, new StreamResult(out));
									data = out.toByteArray();
								} else if (mode.equals("dom")) {
									ByteArrayOutputStream out = new ByteArrayOutputStream();
									domSerializer.transform(new DOMSource(domDoc), new StreamResult(out));
									data = out.toByteArray();
								} else if (mode.startsWith("fi")) {
									data = serializeWithFastInfoset(doc, fiSerializer, fiMethod);
								} else {
									throw new IllegalArgumentException("illegal mode");
								}
								checksum += data.length;
							}
							encodedSize = data.length;
							doneEncoded += encodedSize;
							
							// deserialize
							Document doc2 = null;
							if (cmd.equals("deser") || cmd.equals("serdeser") || cmd.equals("test")) {
								if (mode.startsWith("bnux")) {
									doc2 = codec.deserialize(new ByteArrayInputStream(data), bnuxFactory);
								} else if (mode.startsWith("xom") && mode.indexOf("stax") >= 0) { 
									doc2 = staxBuilder.build(new ByteArrayInputStream(fileData));							
								} else if (mode.startsWith("xom")) { 
									if (mode.indexOf("stream") < 0) {
										doc2 = builder.build(new ByteArrayInputStream(fileData), baseURI);
									} else {
										doc2 = builder.build(new ByteArrayInputStream(data), baseURI);
									}
								} else if (mode.equals("saxon")) { // just for deser comparison
									context.buildDocument(new StreamSource(new ByteArrayInputStream(fileData)));
								} else if (mode.equals("dom")) {
									domDoc = domBuilder.parse(new ByteArrayInputStream(fileData));
//									System.err.println(domDoc.getClass().getName());
								} else if (mode.startsWith("fi") && mode.indexOf("stax") >=0 ) {
									fistaxMethod.invoke(fistaxReader, new Object[] {new ByteArrayInputStream(data)});
									doc2 = staxBuilder.build(fistaxReader);
								} else if (mode.startsWith("fi")) {
									doc2 = fiBuilder.build(new ByteArrayInputStream(data));
								} else {
									throw new IllegalArgumentException("illegal mode");
								}							
								if (doc2 != null) checksum += doc2.getBaseURI().length();
							}
							
							// check correctness
							if (cmd.equals("test")) {
								if (! Arrays.equals(XOMUtil.toCanonicalXML(doc), XOMUtil.toCanonicalXML(doc2))) {
									System.err.println("Canonical XML Mismatch: ");
									System.err.println("expected: " + doc.toXML());
									System.err.println("actual: " + doc2.toXML());									
									printDiff(doc, doc2);
									System.exit(0);
								}
								if (!equalsDocTypeEquals(doc.getDocType(), doc2.getDocType())) {	
									System.err.println("DocType Mismatch: ");
									System.err.println("expected: " + doc.toXML());
									System.err.println("actual: " + doc2.toXML());
									System.exit(0);
								}
							}
						} catch (RuntimeException e) {
							System.err.println("FATAL ERROR: " + e);
							System.err.println("expected" + doc.toXML());
							throw e;
						}
						done += fileLength;
					}
					time += (System.currentTimeMillis() - start);
					float cfactor = fileLength * 1.0f / encodedSize;
					System.out.println("compression factor = " + cfactor);			
					if (testCompressionLevels) // alternate on each iteration
						compressionLevel = (compressionLevel + 1) % 10;
				} catch (Exception e) {
					if (hasCause(e, "org.jvnet.fastinfoset.FastInfosetException")) {
						; // FI bug; account as 0 MB/s and continue unbothered
						e.printStackTrace(System.out);
					} else {
						throw e;
					}
				}
			}
					
			
			System.out.println("\n****** SUMMARY ******");
			System.out.println("files = " + iterations * (args.length - 5));
			System.out.println("secs = " + (time / 1000.0f));
			System.out.println("mean throughput MB/s = "
					+ ((done / (1024.0f * 1024.0f))
					/ (time / 1000.0f)));
			float cfactor = done * 1.0f / doneEncoded;
			System.out.println("mean compression factor = " + cfactor);		
			System.out.println("files/sec = " + (iterations * (args.length - 5) / (time / 1000.0f)));		
			System.out.println("checksum = " + checksum);	
		}		
	}
	
	private static byte[] serializeWithXOM(Document doc) throws IOException {
//		return doc.toXML().getBytes();
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		Serializer ser = new Serializer(out);
//		ser.setIndent(4);
		ser.write(doc);
		return out.toByteArray();
	}
	
	private static byte[] serializeWithStreamingXOM(Document doc) throws IOException {
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		StreamingSerializerFactory factory = new StreamingSerializerFactory();
		StreamingSerializer ser = factory.createXMLSerializer(out, "UTF-8");
//		ser.setIndent(4);
		ser.write(doc);
		return out.toByteArray();
	}
	
	private static byte[] serializeWithStreamingBnux(Document doc, int compressionLevel) throws IOException {
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		StreamingSerializerFactory factory = new StreamingSerializerFactory();
		StreamingSerializer ser = factory.createBinaryXMLSerializer(out, compressionLevel);
		ser.write(doc);
		return out.toByteArray();
	}
	
	private static byte[] serializeWithFastInfoset(Document doc, ContentHandler fiSerializer, java.lang.reflect.Method method) throws Exception {
		// work-around to avoid making FastInfoSet.jar a hard dependency.
		// for best performance on very small documents reflection should better not be used...
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		try {
			method.invoke(fiSerializer, new Object[] {out});
		} catch (Throwable t) {
			throw new Error(t);
		}
		new SAXConverter(fiSerializer).convert(doc);
		return out.toByteArray();
	}
	
	private static boolean hasCause(Throwable e, String clazz) {
		while (e != null) {
//			System.out.println("chain="+ e.getClass().getName());
			if (e.getClass().getName().equals(clazz)) return true;
			// walk exception cause chain:
			if (e instanceof SAXException) {
				e = ((SAXException)e).getException();
			} else if (e instanceof XMLStreamException) {
				e = ((XMLStreamException)e).getNestedException();
			} else {
				e = e.getCause();
			}
		}
		return false;
	}
	
	// print snippet of the offending area to gain some debugging clues
	private static void printDiff(Document expected, Document actual) {
		int window = 100;
//		int window = 300;
//		int window = 10;
		System.err.println("Canonical XML Diff Location Snippet:");
		byte[] e = XOMUtil.toCanonicalXML(expected);
		byte[] a = XOMUtil.toCanonicalXML(actual);
		if (e.length != a.length) {
			System.err.println("e.length="+ e.length + ", a.length=" + a.length);
		}
		int size = Math.min(e.length, a.length);
		for (int i=0; i < size; i++) {
			if (e[i] != a[i]) {
				System.err.println("diff at i=" + i + ", e[i]=" + e[i] + 
						", a[i]=" + a[i]);
				int off = Math.max(0, i-window);
//				int off = Math.max(0, i-0);
				int len = Math.min(2*window, size-off);
				try {
					System.err.println("e='"+ new String(e, off, len, "UTF-8") + "'");
					System.err.println("a='"+ new String(a, off, len, "UTF-8") + "'");
//					System.err.println("eb='"+ new ArrayByteList(new String(e, off, len, "UTF-8").getBytes("UTF-8")));
//					System.err.println("ab='"+ new ArrayByteList(new String(a, off, len, "UTF-8").getBytes("UTF-8")));
				} catch (UnsupportedEncodingException ex) {
					throw new RuntimeException("can never happen");
				}
				break;
			}
		}
	}
	
//	private static void parseDOM(File file) throws Exception {
//		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
//		DocumentBuilder domBuilder = factory.newDocumentBuilder();
//		org.w3c.dom.Document dom = domBuilder.parse(file);
////		System.out.println(dom.getClass());
//
//		System.setProperty("javax.xml.transform.TransformerFactory",
//				"org.apache.xalan.processor.TransformerFactoryImpl");
//		TransformerFactory transFactory = TransformerFactory.newInstance();
//		Transformer idTransform = transFactory.newTransformer();
//		ByteArrayOutputStream out = new ByteArrayOutputStream();
//		idTransform.transform(new DOMSource(dom), new StreamResult(out));
//		System.out.println(out.toString());
//	}
	
	private static boolean equalsDocTypeEquals(DocType x, DocType y) {
		if (x != y) {
			if (x == null && y != null)
				return false;
			if (x != null && y == null)
				return false;
			
			if (!x.getInternalDTDSubset().equals(y.getInternalDTDSubset()))
				return false;
			if (!eq(x.getPublicID(), y.getPublicID()))
				return false;
			if (!eq(x.getSystemID(), y.getSystemID()))
				return false;
			if (!eq(x.getRootElementName(), y.getRootElementName()))
				return false;
			return true;
		}

		return true;
	}
	
	private static boolean eq(Object x, Object y) {
		if (x != y) {
			if (x == null && y != null)
				return false;
			if (x != null && y == null)
				return false;
			if (!x.equals(y))
				return false;
		}

		return true;
	}
	
	// ignore some stuff from the test suite dirs
	private static boolean ignore(String file) {
		if (file.endsWith(".html")) return true;
		if (file.endsWith(".out")) return true;
		if (file.endsWith(".txt")) return true;
		if (file.endsWith(".bat")) return true;
		if (file.endsWith(".sh")) return true;
		if (file.endsWith(".sxx")) return true;
		if (file.endsWith(".ssx")) return true;
		if (file.endsWith(".dtd")) return true;
		if (file.endsWith(".xq")) return true;
		if (file.endsWith(".zip")) return true;
		if (file.endsWith(".gz")) return true;
		if (file.endsWith(".jpg")) return true;
		if (file.endsWith(".gif")) return true;
//		if (file.endsWith("atrs13.xml")) return true; // char encoding
//		if (file.endsWith("atrs15.xml")) return true; // char encoding
		
//		if (file.endsWith("atrs47.xsl")) return true;
//		if (file.endsWith("atrs48.xsl")) return true;
//		if (file.endsWith("idky04.xml")) return true;
//		if (file.endsWith("idky10.xml")) return true;
//		if (file.endsWith("idky104.xml")) return true;
//		if (file.endsWith("idky11.xml")) return true;
//		if (file.endsWith("idky122.xml")) return true;
//		if (file.endsWith("idky123.xml")) return true;
		
		if (file.endsWith("nspc05x.xml")) return true; // not wellformed
//		if (file.endsWith("idky28.xml")) return true; // char encoding
//		if (file.endsWith("idky28wolf.xml")) return true; // char encoding
//		if (file.endsWith("idky35.xml")) return true; // char encoding
//		if (file.endsWith("stdxmlfile.xml")) return true; // char encoding
//		if (file.endsWith("copy19.xsl")) return true; // char encoding
//		if (file.endsWith("copy20.xsl")) return true; // char encoding
//		if (file.endsWith("copy22.xml")) return true; // char encoding
//		if (file.endsWith("copy38.xsl")) return true; // char encoding
		if (file.endsWith("ent22.xml")) return true; // not wellformed
		if (file.endsWith("2")) return true; // not wellformed

//		if (file.endsWith("expr85.xml")) return true; // char encoding
//		if (file.endsWith("numberformat06.xsl")) return true; // char encoding
//		if (file.endsWith("numberformat07.xsl")) return true; // char encoding
//		if (file.endsWith("str122.xsl")) return true; // char encoding
//		if (file.endsWith("str126.xsl")) return true; // char encoding
//		if (file.endsWith("str127.xsl")) return true; // char encoding
//		
//		if (file.indexOf("entref") >= 0) return true; // char encoding
		
		if (file.indexOf("err.") >= 0) return true;
		if (file.indexOf("ERR.") >= 0) return true;
//		if (file.indexOf("fail") >= 0) return true;
		if (file.indexOf("Untitled") >= 0) return true;
		
		// xom test suite:
		if (file.endsWith("CVS")) return true; // CVS messes up char encoding
		if (file.endsWith("ucs4bigendian.xml")) return true; // CVS messes up char encoding
		if (file.endsWith("ucs4littleendian.xml")) return true; // CVS messes up char encoding
		if (file.endsWith("utf8.xml")) return true; // CVS messes up char encoding
		if (file.endsWith("097.ent")) return true; // CVS messes up char encoding
		if (file.endsWith("test_ebcdic.xml")) return true; // CVS messes up char encoding
		
		// W3C XML test suite:
		if (file.endsWith(".ent")) return true; 
		if (file.endsWith("xmltest/valid/sa/012.xml")) return true; // invalid
		if (file.indexOf("fail") >= 0) return true;
		if (file.endsWith("p04pass1.xml")) return true; 
		if (file.endsWith("namespaces/1.0/009.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/010.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/011.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/013.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/014.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/015.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/016.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/023.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/025.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/026.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/029.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.0/030.xml")) return true; // not wellformed
		if (file.endsWith("namespaces/1.1/005.xml")) return true; // not wellformed
		
//		if (file.endsWith("valid/sa/066.xml")) return true; // xerces-2.6.2 internal subset
//		if (file.endsWith("valid/sa/101.xml")) return true; // xerces-2.6.2 internal subset
//		if (file.endsWith("valid/sa/090.xml")) return true; // crimson internal subset
		
		if (file.endsWith("xinclude/input/utf8bom.xml")) return true; // crimson can't handle BOMs
		if (file.endsWith("wolf-xom/dtdtest.xhtml")) return true; // crimson bug
		if (file.endsWith("wolf-xom/xinclude/output/langtest3.xml")) return true; // crimson bug
		if (file.indexOf("wolf/large/svg/tests/spec/color/") >= 0) return true; // more crimson bugs
		if (file.indexOf("wolf/large/svg/tests/spec/coordinates/") >= 0) return true;
		if (file.indexOf("wolf/large/svg/tests/spec/filters/") >= 0) return true;
		if (file.indexOf("wolf/large/svg/tests/spec/fonts/") >= 0) return true;
		
		return false;
	}
		
}