/*
 * Copyright (c) 2005, The Regents of the University of California, through
 * Lawrence Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy). All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * (2) Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * (3) Neither the name of the University of California, Lawrence Berkeley
 * National Laboratory, U.S. Dept. of Energy nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * You are under no obligation whatsoever to provide any bug fixes, patches, or
 * upgrades to the features, functionality or performance of the source code
 * ("Enhancements") to anyone; however, if you choose to make your Enhancements
 * available either publicly, or directly to Lawrence Berkeley National
 * Laboratory, without imposing a separate written license agreement for such
 * Enhancements, then you hereby grant the following license: a non-exclusive,
 * royalty-free perpetual license to install, use, modify, prepare derivative
 * works, incorporate into other computer software, distribute, and sublicense
 * such enhancements or derivative works thereof, in binary and source code
 * form.
 */
package nux.xom.tests;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.nio.charset.Charset;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Node;
import nu.xom.Nodes;
import nux.xom.pool.BuilderPool;
import nux.xom.pool.FileUtil;
import nux.xom.pool.XOMUtil;
import nux.xom.xquery.ResultSequenceSerializer;
import nux.xom.xquery.XQuery;
import nux.xom.xquery.XQueryUtil;

/**
 * Runs the <a target="_blank" href="http://www.w3.org/XML/Query/test-suite/">
 * Official W3C XQuery Test Suite</a> (XQTS) against Nux, looking for potential
 * standards conformance bugs; The test suite contains some 8000 test cases; it
 * must be downloaded separately from the W3C site.
 * <p>
 * Example usage:
 * 
 * <pre>
 * export CLASSPATH=lib/nux.jar:lib/saxon8.jar:lib/xom.jar
 * java nux.xom.tests.XQueryTestSuiteW3C ../xqts-0.8.0
 * </pre>
 * 
 * @author whoschek.AT.lbl.DOT.gov
 * @author $Author: hoschek3 $
 * @version $Revision: 1.10 $, $Date: 2005/11/28 21:04:19 $
 */
public class XQueryTestSuiteW3C {

	private static final Charset UTF8 = Charset.forName("UTF-8");
	
	private XQueryTestSuiteW3C() {}
	
	/**
	 * Runs the test suite; the first argument indicates the test suite's root
	 * directory.
	 */
	public static void main(String[] args) throws Throwable {
		new XQueryTestSuiteW3C().run(args);
	}
	
	private void run(String[] args) throws Throwable {
		if (args.length == 0) args = new String[] { "../xqts-0.8.0" };
		File rootDir = new File(args[0]);
		if (!rootDir.exists() || !rootDir.isDirectory()) {
			throw new IllegalArgumentException("xqts dir does not exist: " + rootDir);
		}
		Document catalog = buildDocument(new File(rootDir, "XQTSCatalog.xml"));
		
		String ns = "declare namespace ns = 'http://www.w3.org/2005/02/query-test-XQTSCatalog'; ";
//		String version = XQueryUtil.xquery(catalog, ns + "ns:test-suite/@version").get(0).getValue();
		File expectedDir = new File(rootDir, 
			XQueryUtil.xquery(catalog, ns + "ns:test-suite/@ResultOffsetPath").get(0).getValue());
		File queryDir = new File(rootDir, 
			XQueryUtil.xquery(catalog, ns + "ns:test-suite/@XQueryQueryOffsetPath").get(0).getValue());
		File testSourcesDir = new File(rootDir, "TestSources");		
		Nodes testCases = XQueryUtil.xquery(catalog, ns + "//ns:test-case");
		
		for (int i=0; i < testCases.size(); i++) {
			Node testCase = testCases.get(i);
			String path = XQueryUtil.xquery(testCase, "@FilePath").get(0).getValue();
			File query = new File(new File(queryDir, path), 
				XQueryUtil.xquery(testCase, ns + "ns:query/@name").get(0).getValue() + ".xq");
			String squery = readQuery(query); 
			System.out.println(i + ": " + query + " ...");
			if (squery == null) {
				System.out.println("		************* IGNORED *****");
				continue;
			}
						
			Nodes inputs = XQueryUtil.xquery(testCase, ns + "ns:input-file");
			Map vars = new HashMap();
			for (int j=0; j < inputs.size(); j++) {
				File input = new File(testSourcesDir, inputs.get(j).getValue() + ".xml");
				String varName = ((Element) inputs.get(j)).getAttributeValue("variable");
				Document inputDoc = buildDocument(input);
				if (true) XOMUtil.Normalizer.STRIP.normalize(inputDoc);
				vars.put(varName, inputDoc);
			}

			Nodes expectedErrors = XQueryUtil.xquery(testCase, ns + "ns:expected-error");					
			Nodes results = null;
			try { // here's where the query is actually executed
				XQuery xquery = new XQuery(squery, query.toURI());
				results = xquery.execute(null, null, vars).toNodes();
			} catch (Throwable t) {
				if (expectedErrors.size() == 0) {
					System.out.println(XOMUtil.toPrettyXML(testCase));
					throw t;
				}
//				System.out.println("expected error:" + t);
				continue;
			}
			
			Nodes expectedOutputs = XQueryUtil.xquery(testCase, ns + "ns:output-file");
			for (int k=0; k < expectedOutputs.size(); k++) {
				File expectedOutput = new File(
						new File(expectedDir, path), expectedOutputs.get(k).getValue());			
				String compare = ((Element)expectedOutputs.get(k)).getAttributeValue("compare");
				try {
					if (compare.equals("Text")) {
						String expected = FileUtil.toString(
								new FileInputStream(expectedOutput), UTF8);
						String actual = serialize(results);
						assertEquals(expected, actual);
					} else if (compare.equals("XML")) {
						Document expected = buildDocument(expectedOutput);
						if (query.toString().indexOf("XQuery/UseCase/") >= 0) {
							// input doc should not have whitespace
							XOMUtil.Normalizer.STRIP.normalize(expected); // ???
						}
						Document actual = XOMUtil.toDocument(serialize(results));
						assertEquals(expected, actual);				
					} else if (compare.equals("Fragment")) {				
						Document expected = buildFromSequence(expectedOutput);
						Document actual = buildFromSequence(results);
						assertEquals(expected, actual);
					} else if (compare.equals("Ignore")) {
						; // nothing to do
					} else if (compare.equals("Inspect")) { 
						System.out.println("****************** Inspect output?");
					} else {
						throw new RuntimeException(
							"Unrecognized comparison operator: " + compare);
					}			
					break; // found a match; break out of "for" loop
				} catch (ConformanceException e) {
					if (k == expectedOutputs.size()-1) { // done trying all expected outputs?
						System.out.println(XOMUtil.toPrettyXML(testCase));
						throw e;				
					}
				}
			}		
//			System.out.println("Passed.");
		}
		System.out.println("\nFinished testing. Good bye.");
	}
	
	private void assertEquals(String expected, String actual) {
		if (!normalize(expected).equals(normalize(actual))) {
//			fail(new Text(expected), new Text(actual));
			fail(expected, actual);
		}
	}
	
	private void assertEquals(Document expected, Document actual) throws UnsupportedEncodingException {
		int window = 20;
		byte[] e = XOMUtil.toCanonicalXML(expected);
		byte[] a = XOMUtil.toCanonicalXML(actual);
		if (!Arrays.equals(e, a)) {
			// print snippet of the offending area to gain some debugging clues
			if (e.length != a.length) {
				System.out.println("e.length="+ e.length + ", a.length=" + a.length);
			}
			int size = Math.min(e.length, a.length);
			for (int i=0; i < size; i++) {
				if (e[i] != a[i]) {
					System.out.println("diff at i=" + i + ", e[i]=" + e[i] + 
							", a[i]=" + a[i]);
					int off = Math.max(0, i-window);
					int len1 = Math.min(2*window, e.length-off);
					int len2 = Math.min(2*window, a.length-off);
					System.out.println("e='"+ new String(e, off, len1, "UTF-8") + "'");
					System.out.println("a='"+ new String(a, off, len2, "UTF-8") + "'");
//					System.out.println("e1='"+ new String(e, "UTF-8") + "'");
//					System.out.println("a1='"+ new String(a, "UTF-8") + "'");
					break;
				}
			}
//			fail(expected, actual);
			fail(expected.toXML(), actual.toXML());
		}
	}
	
	private void fail(Object expected, Object actual) {
		throw new ConformanceException(
			"\nexpected='" + expected + "', \nactual  ='" + actual + "'");		
	}
	
	private String normalize(String text) {
		Element wrapper = new Element("dummy");
		wrapper.appendChild(text);
		XOMUtil.Normalizer.COLLAPSE.normalize(wrapper); // ???
		return wrapper.getValue();
	}
	
	private Document buildDocument(File file) throws Exception {
		return BuilderPool.GLOBAL_POOL.getBuilder(false).build(file);
	}
	
	private Document buildFromSequence(File file) {
		String xml = "<!DOCTYPE doc [<!ENTITY e SYSTEM '" + file.toURI() + 
				"'>]><doc>&e;</doc>";
		Document doc = XOMUtil.toDocument(xml);
		if (true) XOMUtil.Normalizer.COLLAPSE.normalize(doc); // needed by SeqUnion/fn-union-node-args-004.xq et al
		return doc;
	}
	
	private Document buildFromSequence(Nodes nodes) throws IOException {
		File tmp = new File(System.getProperty("java.io.tmpdir"));
		tmp = new File(tmp, "xqts-tmp.out");
		Writer out = new OutputStreamWriter(new FileOutputStream(tmp), UTF8);
		out.write(serialize(nodes));
		out.flush();
		out.close();	
		Document doc = buildFromSequence(tmp); // reparse
		tmp.delete();
		return doc;
	}
	
	private String serialize(Nodes nodes) {
		ResultSequenceSerializer serializer = new ResultSequenceSerializer();
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		String xml;
		try {
			serializer.write(nodes, out);
			xml = out.toString("UTF-8"); // safe: UTF-8 support is required by JDK spec
		} catch (IOException e) {
			throw new RuntimeException("should never happen", e);
		}
		
		// remove XML declaration header <?xml version="1.0" encoding="UTF-8"?>\r\n
		// remove trailing line break, if any
		xml = xml.substring(xml.indexOf('>') + 1);
		if (xml.startsWith("\r\n")) xml = xml.substring(2);
		int j = xml.length();
		if (xml.endsWith("\r\n")) j = j - 2;
		else if (xml.endsWith("\n")) j = j - 1;
		return xml.substring(0, j);
	}

	private String readQuery(File query) throws IOException {
		if (ignore(query)) return null;		
		String squery = FileUtil.toString(new FileInputStream(query), UTF8);	
		return substituteSchema(squery);		
	}
	
	/** Nux is not schema aware: replace schema refs as suggested by W3C */
	private String substituteSchema(String query) {
		String toFind = "import schema namespace";
		int i = query.indexOf(toFind);
		if (i >= 0) { 
			System.out.println("************* Substituting 'import schema namespace' with 'declare namespace'.");
			System.out.println("before='" + query + "'");
			query = query.substring(0, i) + "declare namespace" + query.substring(i + toFind.length());
			System.out.println("after='" + query + "'");
		}
		
		toFind = "import schema";
		i = query.indexOf(toFind);
		if (i >= 0) {
			System.out.println("************* Substituting 'import schema' with 'declare default element namespace'.");
			System.out.println("before='" + query + "'");
			if (query.indexOf("declare default element namespace") < 0) {
				query = query.substring(0, i) + "declare default element namespace" + query.substring(i + toFind.length());
			} else { // there is already a default element namespace; remove import schema "xyz"; 
				int j = query.indexOf(';', i);
				query = query.substring(0, i) + query.substring(j+1);
			}
			System.out.println("after='" + query + "'");
		}
		return query;
	}
	
	private boolean ignore(File query) {
		String s = query.toString();
//		if (s.indexOf("UseCaseXMP") < 0) return true;
		
		if (s.endsWith("SeqUnion/fn-union-node-args-015.xq")) return true; // W3C bug (missing inscope namespaces)		
		if (s.endsWith("SeqUnion/fn-union-node-args-016.xq")) return true; // W3C bug (missing inscope namespaces)					
		if (s.endsWith("SeqUnion/fn-union-node-args-017.xq")) return true; // W3C bug (missing inscope namespaces)	
		if (s.endsWith("SeqUnion/fn-union-node-args-020.xq")) return true; // W3C bug (missing inscope namespaces)	
		if (s.endsWith("SeqUnion/fn-union-node-args-023.xq")) return true; // missing whitespace in xqts-0.8.0 file
		
		if (s.endsWith("SeqIntersect/fn-intersect-node-args-014.xq")) return true; // W3C bug (missing inscope namespaces)	
		if (s.endsWith("SeqIntersect/fn-intersect-node-args-015.xq")) return true; // W3C bug (missing inscope namespaces)
		if (s.endsWith("SeqIntersect/fn-intersect-node-args-016.xq")) return true; // W3C bug (missing inscope namespaces)					
		if (s.endsWith("SeqIntersect/fn-intersect-node-args-017.xq")) return true; // W3C bug (missing inscope namespaces)
		if (s.endsWith("SeqIntersect/fn-intersect-node-args-023.xq")) return true; // missing whitespace in xqts-0.8.0 file	
		
		if (s.endsWith("SeqExcept/fn-except-node-args-008.xq")) return true; // missing whitespace in xqts-0.8.0 file
		if (s.endsWith("SeqExcept/fn-except-node-args-017.xq")) return true; // W3C bug (missing inscope namespaces)
		if (s.endsWith("SeqExcept/fn-except-node-args-022.xq")) return true; // missing whitespace in xqts-0.8.0 file
				
		if (s.indexOf("DayTimeDurationDivideDTD") >= 0) return true; // same pattern 
				
		if (s.endsWith("DirectConElemContent/Constr-cont-constrmod-2.xq")) return true; // W3C bug
		if (s.endsWith("DirectConElemContent/Constr-cont-constrmod-4.xq")) return true; // double -> int
		if (s.endsWith("DirectConElemContent/Constr-cont-constrmod-6.xq")) return true; // W3C bug
		if (s.endsWith("DirectConElemContent/Constr-cont-constrmod-8.xq")) return true; // double -> int		
		if (s.endsWith("DirectConElemContent/Constr-cont-nsmode-4.xq")) return true; // DefaultResultSequence ns redefinition bug???
			
		if (s.endsWith("ComputeConElem/Constr-compelem-compname-9.xq")) return true; // W3C bug: undeclared prefix
		if (s.endsWith("ComputeConElem/Constr-compelem-compname-13.xq")) return true; // W3C bug: ns moved up
		if (s.endsWith("ComputeConElem/Constr-compelem-compname-16.xq")) return true; // W3C bug: ns moved up
		if (s.endsWith("ComputeConElem/Constr-compelem-constrmod-2.xq")) return true; // text normalization bug
		if (s.endsWith("ComputeConElem/Constr-compelem-constrmod-4.xq")) return true; // double -> int
		if (s.endsWith("ComputeConElem/Constr-compelem-constrmod-6.xq")) return true; // W3C bug
		if (s.endsWith("ComputeConElem/Constr-compelem-constrmod-8.xq")) return true; // double -> int
		
		if (s.endsWith("ComputeConAttr/Constr-compattr-compname-9.xq")) return true; // // W3C bug: undeclared prefix
		
		// XPath -> XOM data model mismatches:
		if (s.endsWith("ConDocNode/Constr-docnode-nested-1.xq")) return true; // Cannot add a nu.xom.Text to a Document
		if (s.endsWith("ConDocNode/Constr-docnode-enclexpr-1.xq")) return true; // Missing document root element; A XOM document must have a root element
		if (s.endsWith("ConDocNode/Constr-docnode-enclexpr-2.xq")) return true; // Cannot add a nu.xom.Text to a Document
		if (s.endsWith("ConDocNode/Constr-docnode-enclexpr-3.xq")) return true; // Cannot add a nu.xom.Text to a Document
		if (s.endsWith("ConDocNode/Constr-docnode-enclexpr-4.xq")) return true; // Cannot add a nu.xom.Text to a Document
		if (s.endsWith("ConDocNode/Constr-docnode-doc-1.xq")) return true; // A XOM document must not have more than one root element
		if (s.endsWith("ConText/Constr-text-nested-3.xq")) return true; // Cannot add a nu.xom.Text to a Document
	
		if (s.endsWith("ConDocNode/Constr-docnode-constrmod-2.xq")) return true; // W3C bug
		if (s.endsWith("ConDocNode/Constr-docnode-constrmod-4.xq")) return true; // double -> int
		if (s.endsWith("ConDocNode/Constr-docnode-constrmod-6.xq")) return true; // W3C bug 
		if (s.endsWith("ConDocNode/Constr-docnode-constrmod-4.xq")) return true; // double -> int
			
		if (s.endsWith("ComputeConComment/Constr-compcomment-nested-3.xq")) return true; // A XOM document must not have more than one root element
		if (s.endsWith("ComputeConComment/Constr-compcomment-dash-4.xq")) return true; // file not found (should be emptydoc)
		if (s.endsWith("ComputeConComment/Constr-compcomment-doubledash-4.xq")) return true; // file not found (should be emptydoc)
		
		if (s.endsWith("OrderbyExprWith/orderBy20.xq")) return true; // W3C bug: descending vs. ascending
		if (s.endsWith("OrderbyExprWith/orderBy21.xq")) return true; // number formatting????
		if (s.endsWith("OrderbyExprWith/orderBy22.xq")) return true; // number formatting????
		if (s.endsWith("OrderbyExprWith/orderBy25.xq")) return true; // number formatting????
		if (s.endsWith("OrderbyExprWith/orderBy27.xq")) return true; // number formatting????
		if (s.endsWith("OrderbyExprWith/orderBy29.xq")) return true; // number formatting: 100000000000000000 vs. 1.0E17
		if (s.endsWith("OrderbyExprWith/orderBy32.xq")) return true; // number formatting????
		if (s.endsWith("OrderbyExprWith/orderBy35.xq")) return true; // number formatting????
		if (s.endsWith("OrderbyExprWith/orderBy42.xq")) return true; // number formatting????
		if (s.endsWith("OrderbyExprWith/orderBy45.xq")) return true; // number formatting????
		if (s.endsWith("OrderbyExprWith/orderBy50.xq")) return true; // number formatting????
		if (s.endsWith("OrderbyExprWith/orderBy51.xq")) return true; // number formatting????
		if (s.endsWith("OrderbyExprWith/orderBy52.xq")) return true; // number formatting????
		if (s.endsWith("OrderbyExprWith/orderBy55.xq")) return true; // number formatting????
		if (s.endsWith("OrderbyExprWith/orderBy56.xq")) return true; // number formatting: 0 vs. -0 
		if (s.endsWith("OrderbyExprWith/orderBy57.xq")) return true; // number formatting: 0 vs. -0 
		
		if (s.endsWith("SeqExprCast/casthcds3.xq")) return true; // string -> boolean
		if (s.endsWith("SeqExprCast/casthcds9.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds10.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds11.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds13.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds15.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds20.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds21.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds27.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds30.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds31.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds33.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds35.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds36.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds38.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds39.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds40.xq")) return true; // number formatting
		if (s.endsWith("SeqExprCast/casthcds41.xq")) return true; // number formatting		
		
		if (s.endsWith("InternalVariablesWithout/VarDecl009.xq")) return true; // "<" escaping
		if (s.endsWith("UseCaseNS/ns-queries-results-q7.xq")) return true; // Saxon or ResultSeqSerializer or DefaultResSeq introduce unncecessary namespace declaration (taken from inscope ns)

		if (s.endsWith("Catalog/Catalog001.xq")) return true; // file not found (should be ../xyz)
		if (s.endsWith("Catalog/Catalog002.xq")) return true; // file not found (should be ../xyz)
		if (s.endsWith("Catalog/Catalog003.xq")) return true; // file not found (should be ../xyz)
		
		
		// saxon-8.6:
		if (s.indexOf("YearMonthDurationDivideYMD/op-divide-yearMonthDuration-by-yMD-11.xq") >= 0) return true;
		if (s.indexOf("DurationDateTimeOp/TimeGT/op-time-greater-than-2.xq") >= 0) return true; 
		if (s.indexOf("DurationDateTimeOp/gYearMonthEQ/op-gYearMonth-equalNew-2.xq") >= 0) return true; 
		if (s.indexOf("TimezoneFunction/AdjDateToTimezoneFunc/fn-adjust-date-to-timezone-18.xq") >= 0) return true; 
		if (s.indexOf("TimezoneFunction/AdjDateToTimezoneFunc/fn-adjust-date-to-timezone-19.xq") >= 0) return true;
		if (s.indexOf("TimezoneFunction/AdjTimeToTimezoneFunc/fn-adjust-time-to-timezone-18.xq") >= 0) return true;
		if (s.indexOf("TimezoneFunction/AdjTimeToTimezoneFunc/fn-adjust-time-to-timezone-19.xq") >= 0) return true;
		
		// xqts-0.8.0:
		if (s.indexOf("PrimaryExpr/Literals057.xq") >= 0) return true; 
		if (s.indexOf("PrimaryExpr/Literals058.xq") >= 0) return true; 
		if (s.indexOf("PrimaryExpr/Literals062.xq") >= 0) return true; 
		if (s.indexOf("PrimaryExpr/Literals") >= 0) return true; 
		if (s.indexOf("PathExpr/Predicates") >= 0) return true; 
			
		if (s.indexOf("ForExprType") >= 0) return true; 
		if (s.indexOf("ForExprType/ForExprType009.xq") >= 0) return true; 
		if (s.indexOf("ForExprType/ForExprType010.xq") >= 0) return true; 
		
		if (s.indexOf("OrderExpr/Orderexpr-15") >= 0) return true; // W3C bug
		if (s.indexOf("OrderExpr/Orderexpr-16") >= 0) return true; // W3C bug
		if (s.indexOf("OrderExpr/Orderexpr-17") >= 0) return true; // W3C bug
		if (s.indexOf("OrderExpr/Orderexpr-18") >= 0) return true; // W3C bug
		
		if (s.indexOf("BoundarySpace") >= 0) return true; // W3C bug
		if (s.indexOf("DefaultOrderingProlog") >= 0) return true;
		if (s.indexOf("CodepointToStringFunc") >= 0) return true;
		if (s.indexOf("TranslateFunc") >= 0) return true;
		if (s.indexOf("EscapingFuncs") >= 0) return true;
		if (s.indexOf("MatchStringFunc") >= 0) return true;
		if (s.indexOf("QNameFunc") >= 0) return true;
		if (s.indexOf("NodeLangFunc") >= 0) return true;

		// saxonb-8.6.1
//		if (s.endsWith("UseCaseXMP/xmp-queries-results-q12.xq")) return true; // saxon deep-equals NPE	
		
		// needed by sun-jdk-1.5.0-xerces-internal bug (begin/end entity)
		// bug appears not to be present in xerces-2.7.1 and crimson 
//		if (s.endsWith("DirectConOther/Constr-pi-content-3.xq")) return true;
//		if (s.endsWith("ComputeConPI/Constr-comppi-empty-1.xq")) return true;
//		if (s.endsWith("ComputeConPI/Constr-comppi-empty-2.xq")) return true;	
		
		return false;
	}
	
	private static final class ConformanceException extends RuntimeException {
		private ConformanceException(String msg) { 
			super(msg); 
		}
	}
	
}
